#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "decoder.h"

_COBHeader	header;
_function	*pFunctions;
_piece		*pPieces;
_label		*pLabels;

long		lnLabels;

extern bool	bStrict;

long printcode(char *psz)
	{
	char szFileName[0x100];
	strcpy(szFileName, psz);
	long lLenFileName = strlen(szFileName);
	szFileName[lLenFileName + 1] = NULL;
	szFileName[lLenFileName] = 'm';
	szFileName[lLenFileName - 1] = 's';
	szFileName[lLenFileName - 2] = 'a';
	szFileName[lLenFileName - 3] = 'b';

	FILE *fp = fopen(szFileName, "w");
	long l = 0;
	lnLabels = 0;
	long z;
	bool bFound;
	bool bJump = false; //only used if bStrict == true

	if (header.lnPieces)
		{
		fprintf(fp, "piece ");
		for (long k = 0; k < header.lnPieces; k++)
			{
			if (k != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%s", pPieces[k].pszName);
			}
		fprintf(fp, ";\n\n");
		}

	if (header.lnStaticVar)
		{
		fprintf(fp, "static-var ");
		for (long m = 0; m < header.lnStaticVar; m++)
			{
			if (m != 0)
				fprintf(fp, ", ");
			fprintf(fp, "static_var_%d", m);
			}
		fprintf(fp, ";\n\n");
		}

	//build the label table
	for (long g = 0; g < header.lnFunctions; g++)
		{
		for (long h = 0; h < pFunctions[g].lSize; h++)
			{
			switch(pFunctions[g].lpCode[h])
				{
				case 0x10064000: //jmp
					bFound = false;
					for (z = 0; (z < lnLabels) && (!bFound); z++)
						{
						if (pLabels[z].lOffset == pFunctions[g].lpCode[h + 1])
							bFound = true;
						}
					if (!bFound)
						{
						lnLabels++;
						pLabels = (_label *)realloc(pLabels, sizeof(_label) * lnLabels);
						pLabels[z].lOffset = pFunctions[g].lpCode[h + 1];
						}
					break;
				case 0x10066000: //jne
					bFound = false;
					for (z = 0; (z < lnLabels) && (!bFound); z++)
						{
						if (pLabels[z].lOffset == pFunctions[g].lpCode[h + 1])
							bFound = true;
						}
					if (!bFound)
						{
						lnLabels++;
						pLabels = (_label *)realloc(pLabels, sizeof(_label) * lnLabels);
						pLabels[z].lOffset = pFunctions[g].lpCode[h + 1];
						}
					break;
				}
			}
		}

	//write the basm code
	long lnLongsSoFar = -1;

	for (long i = 0; i < header.lnFunctions; i++)
		{
		fprintf(fp, "proc %s\t//%d\n\targs none;\n\tvar none;\n\n\tbeginproc\n", pFunctions[i].pszName, i);

		for (long j = 0; j < pFunctions[i].lSize; j++)
			{
			lnLongsSoFar++;
			if (l > 0)
				l--;//skip
			else
				{
				//see if we need to put a label here
				for (long x = 0; x < lnLabels; x++)
					{
					if (pLabels[x].lOffset == lnLongsSoFar)
						{
						fprintf(fp, "\nLabel_%d: //%d\n", x, pLabels[x].lOffset);
						}
					}
				if (bStrict)
					{
					switch(pFunctions[i].lpCode[j])
						{
						case 0x10001000: //move
							fprintf(fp, "\t\tmove;\n");
							break;
						case 0x10002000: //turn
							fprintf(fp, "\t\tturn;\n");
							break;
						case 0x10003000: //spin
							fprintf(fp, "\t\tspin;\n");
							break;
						case 0x10004000: //stop spin
							fprintf(fp, "\t\tstop-spin;\n");
							break;
						case 0x10005000: //show
							fprintf(fp, "\t\tshow;\n");
							break;
						case 0x10006000: //hide
							fprintf(fp, "\t\thide;\n");
							break;
						case 0x10007000: //cache
							fprintf(fp, "\t\tcache;\n");
							break;
						case 0x10008000: //dont cache
							fprintf(fp, "\t\tdont-cache;\n");
							break;
						case 0x10009000: //unk1
							fprintf(fp, "\t\tunk1;\n");
							break;
						case 0x1000A000: //unk2
							fprintf(fp, "\t\tunk2;\n");
							break;
						case 0x1000B000: //move now
							fprintf(fp, "\t\tmovenow;\n");
							break;
						case 0x1000C000: //turn now
							fprintf(fp, "\t\tturnnow;\n");
							break;
						case 0x1000D000: //shade
							fprintf(fp, "\t\tshade;\n");
							break;
						case 0x1000E000: //dont shade
							fprintf(fp, "\t\tdont-shade;\n");
							break;
						case 0x1000F000: //emit sfx
							fprintf(fp, "\t\tsfx;\n");
							break;

						case 0x10011000: //wait turn
							fprintf(fp, "\t\twait-for-turn;\n");
							break;
						case 0x10012000: //wait move
							fprintf(fp, "\t\twait-for-move;\n");
							break;
						case 0x10013000: //wait sleep
							fprintf(fp, "\t\tsleep;\n");
							break;

						case 0x10021000: //unk3
							fprintf(fp, "\t\tunk3;\n");
							break;
						case 0x10021001: //push constant
							fprintf(fp, "\t\tpushc;\n");
							break;
						case 0x10021002: //push var
							fprintf(fp, "\t\tpushl;\n");
							break;
						case 0x10021004: //push static
							fprintf(fp, "\t\tpushs;\n");
							break;
						case 0x10022000: //create local var
							fprintf(fp, "\t\tclv;\n");
							break;
						case 0x10023000: //unk4
							fprintf(fp, "\t\tunk4;\n");
							break;
						case 0x10023002: //pop var
							fprintf(fp, "\t\tpopl;\n");
							break;
						case 0x10023004: //pop static
							fprintf(fp, "\t\tpops;\n");
							break;
						case 0x10024000: //flush
							fprintf(fp, "\t\tflush;\n");
							break;

						case 0x10031000: //math & bitwise instr
							fprintf(fp, "\t\tadd;\n");
							break;
						case 0x10032000:
							fprintf(fp, "\t\tsub;\n");
							break;
						case 0x10033000:
							fprintf(fp, "\t\tmul;\n");
							break;
						case 0x10034000:
							fprintf(fp, "\t\tdiv;\n");
							break;
						case 0x10035000:
							fprintf(fp, "\t\tand;\n");
							break;
						case 0x10036000:
							fprintf(fp, "\t\tor;\n");
							break;
						case 0x10037000:
							fprintf(fp, "\t\txor;\n");
							break;
						case 0x10038000:
							fprintf(fp, "\t\tnot;\n");
							break;

						case 0x10041000: //rand
							fprintf(fp, "\t\trand;\n");
							break;
						case 0x10042000: //get uv
							fprintf(fp, "\t\tgetuv;\n");
							break;
						case 0x10043000: //get
							fprintf(fp, "\t\tget;\n");
							break;
						case 0x10044000: //shl
							fprintf(fp, "\t\tshl;\n");
							break;
						case 0x10045000: //sar
							fprintf(fp, "\t\tsar;\n");
							break;

						case 0x10051000: //logicals
							fprintf(fp, "\t\tsetl;\n");
							break;
						case 0x10052000:
							fprintf(fp, "\t\tsetle;\n");
							break;
						case 0x10053000:
							fprintf(fp, "\t\tsetg;\n");
							break;
						case 0x10054000:
							fprintf(fp, "\t\tsetge;\n");
							break;
						case 0x10055000:
							fprintf(fp, "\t\tcmp;\n");
							break;
						case 0x10056000:
							fprintf(fp, "\t\tsetne;\n");
							break;
						case 0x10057000:
							fprintf(fp, "\t\tland;\n");
							break;
						case 0x10058000:
							fprintf(fp, "\t\tlor;\n");
							break;
						case 0x10059000:
							fprintf(fp, "\t\tlxor;\n");
							break;
						case 0x1005A000:
							fprintf(fp, "\t\tneg;\n");
							break;

						case 0x10061000: //start
							fprintf(fp, "\t\tstart;\n");
							break;
						case 0x10062000: //call
							fprintf(fp, "\t\tcall;\n");
							break;
						case 0x10063000: //unk7
							fprintf(fp, "\t\tunk7\n");
							break;
						case 0x10064000: //jmp
							bJump = true;
							fprintf(fp, "\t\tjmp;\n");
							break;
						case 0x10065000: //ret
							fprintf(fp, "\t\tret;\n");
							break;
						case 0x10066000: //jne
							bJump = true;
							fprintf(fp, "\t\tjne;\n");
							break;
						case 0x10067000: //signal
							fprintf(fp, "\t\tsignal;\n");
							break;
						case 0x10068000: //set signal mask
							fprintf(fp, "\t\tmask;\n");
							break;

						case 0x10071000: //explode
							fprintf(fp, "\t\texplode;\n");	
							break;

						case 0x10082000: //set
							fprintf(fp, "\t\tset;\n");
							break;
						case 0x10083000: //attach
							fprintf(fp, "\t\tattach;\n");
							break;
						case 0x10084000: //drop
							fprintf(fp, "\t\tdrop;\n");
							break;

						case 0x10091000: //cobalt
							fprintf(fp, "\t\t_cobalt;\n");
							break;

						case 0x02800000: //unk8
							fprintf(fp, "\t\tunk8;\n");
							break;

						default: //nop
							if (bJump)
								{
								bJump = false;
								bFound = false;
								for (z = 0; (z < lnLabels) && (!bFound); z++)
									{
									if (pLabels[z].lOffset == pFunctions[i].lpCode[j]) 
										bFound = true; //if the label points to the destination of our jump
									}
								if (!bFound)
									{
									lnLabels++; //we should never get here since we built our label table first
									pLabels = (_label *)realloc(pLabels, sizeof(_label) * lnLabels);
									pLabels[z].lOffset = pFunctions[i].lpCode[j];
									}
								fprintf(fp, "\t\tLabel_%d\n", pLabels[z].lOffset);
								}
							else
								fprintf(fp, "\t\tddw %d;\n", pFunctions[i].lpCode[j]);
							break;
						}
					}
				else
					{
					switch(pFunctions[i].lpCode[j])
						{
						case 0x10001000: //move
							l = 2; //2 post data items
							fprintf(fp, "\t\tmove;\n\t\tddw %d;\n\t\tddw %d;\n", pFunctions[i].lpCode[j + 1], pFunctions[i].lpCode[j + 2]);
							break;
						case 0x10002000: //turn
							l = 2;
							fprintf(fp, "\t\tturn;\n\t\tddw %d;\n\t\tddw %d;\n", pFunctions[i].lpCode[j + 1], pFunctions[i].lpCode[j + 2]);
							break;
						case 0x10003000: //spin
							l = 2;
							fprintf(fp, "\t\tspin;\n\t\tddw %d;\n\t\tddw %d;\n", pFunctions[i].lpCode[j + 1], pFunctions[i].lpCode[j + 2]);
							break;
						case 0x10004000: //stop spin
							l = 2;
							fprintf(fp, "\t\tstop-spin;\n\t\tddw %d;\n\t\tddw %d;\n", pFunctions[i].lpCode[j + 1], pFunctions[i].lpCode[j + 2]);
							break;
						case 0x10005000: //show
							l = 1;
							fprintf(fp, "\t\tshow %d;\n", pFunctions[i].lpCode[j + 1]);
							break;
						case 0x10006000: //hide
							l = 1;
							fprintf(fp, "\t\thide %d;\n", pFunctions[i].lpCode[j + 1]);
							break;
						case 0x10007000: //cache
							l = 1;
							fprintf(fp, "\t\tcache %d;\n", pFunctions[i].lpCode[j + 1]);
							break;
						case 0x10008000: //dont cache
							l = 1;
							fprintf(fp, "\t\tdont-cache %d;\n", pFunctions[i].lpCode[j + 1]);
							break;
						case 0x10009000: //unk1
							break;
						case 0x1000A000: //unk2
							break;
						case 0x1000B000: //move now
							l = 2;
							fprintf(fp, "\t\tmovenow;\n\t\tddw %d;\n\t\tddw %d;\n", pFunctions[i].lpCode[j + 1], pFunctions[i].lpCode[j + 2]);
							break;
						case 0x1000C000: //turn now
							l = 2;
							fprintf(fp, "\t\tturnnow;\n\t\tddw %d;\n\t\tddw %d;\n", pFunctions[i].lpCode[j + 1], pFunctions[i].lpCode[j + 2]);
							break;
						case 0x1000D000: //shade
							l = 1;
							fprintf(fp, "\t\tshade %d;\n", pFunctions[i].lpCode[j + 1]);
							break;
						case 0x1000E000: //dont shade
							l = 1;
							fprintf(fp, "\t\tdont-shade %d;\n", pFunctions[i].lpCode[j + 1]);
							break;
						case 0x1000F000: //emit sfx
							l = 1;
							fprintf(fp, "\t\tsfx %d;\n", pFunctions[i].lpCode[j + 1]);
							break;

						case 0x10011000: //wait turn
							l = 2;
							fprintf(fp, "\t\twait-for-turn;\n\t\tddw %d;\n\t\tddw %d;\n", pFunctions[i].lpCode[j + 1], pFunctions[i].lpCode[j + 2]);
							break;
						case 0x10012000: //wait move
							l = 2;
							fprintf(fp, "\t\twait-for-move;\n\t\tddw %d;\n\t\tddw %d;\n", pFunctions[i].lpCode[j + 1], pFunctions[i].lpCode[j + 2]);
							break;
						case 0x10013000: //wait sleep
							fprintf(fp, "\t\tsleep;\n");
							break;

						case 0x10021000: //unk3
							break;
						case 0x10021001: //push constant
							l = 1;
							fprintf(fp, "\t\tpushc %d;\n", pFunctions[i].lpCode[j + 1]);
							break;
						case 0x10021002: //push var
							l = 1;
							fprintf(fp, "\t\tpushl %d;\n", pFunctions[i].lpCode[j + 1]);
							break;
						case 0x10021004: //push static
							l = 1;
							fprintf(fp, "\t\tpushs static_var_%d;\n", pFunctions[i].lpCode[j + 1]);
							break;
						case 0x10022000: //create local var
							fprintf(fp, "\t\tclv;\n");
							break;
						case 0x10023000: //unk4
							break;
						case 0x10023002: //pop var
							l = 1;
							fprintf(fp, "\t\tpopl %d;\n", pFunctions[i].lpCode[j + 1]);
							break;
						case 0x10023004: //pop static
							l = 1;
							fprintf(fp, "\t\tpops static_var_%d;\n", pFunctions[i].lpCode[j + 1]);
							break;
						case 0x10024000: //flush
							fprintf(fp, "\t\tflush;\n");
							break;

						case 0x10031000: //math & bitwise instr
							fprintf(fp, "\t\tadd;\n");
							break;
						case 0x10032000:
							fprintf(fp, "\t\tsub;\n");
							break;
						case 0x10033000:
							fprintf(fp, "\t\tmul;\n");
							break;
						case 0x10034000:
							fprintf(fp, "\t\tdiv;\n");
							break;
						case 0x10035000:
							fprintf(fp, "\t\tand;\n");
							break;
						case 0x10036000:
							fprintf(fp, "\t\tor;\n");
							break;
						case 0x10037000:
							fprintf(fp, "\t\txor;\n");
							break;
						case 0x10038000:
							fprintf(fp, "\t\tnot;\n");
							break;

						case 0x10041000: //rand
							fprintf(fp, "\t\trand;\n");
							break;
						case 0x10042000: //get uv
							fprintf(fp, "\t\tgetuv;\n");
							break;
						case 0x10043000: //get
							fprintf(fp, "\t\tget;\n");
							break;
						case 0x10044000: //shl
							fprintf(fp, "\t\tshl;\n");
							break;
						case 0x10045000: //sar
							fprintf(fp, "\t\tsar;\n");
							break;

						case 0x10051000: //logicals
							fprintf(fp, "\t\tsetl;\n");
							break;
						case 0x10052000:
							fprintf(fp, "\t\tsetle;\n");
							break;
						case 0x10053000:
							fprintf(fp, "\t\tsetg;\n");
							break;
						case 0x10054000:
							fprintf(fp, "\t\tsetge;\n");
							break;
						case 0x10055000:
							fprintf(fp, "\t\tcmp;\n");
							break;
						case 0x10056000:
							fprintf(fp, "\t\tsetne;\n");
							break;
						case 0x10057000:
							fprintf(fp, "\t\tland;\n");
							break;
						case 0x10058000:
							fprintf(fp, "\t\tlor;\n");
							break;
						case 0x10059000:
							fprintf(fp, "\t\tlxor;\n");
							break;
						case 0x1005A000:
							fprintf(fp, "\t\tneg;\n");
							break;

						case 0x10061000: //start
							l = 2;
							//fprintf(fp, "\t\tstart;\n\t\tddw %d;\n\t\tddw %d;\n", pFunctions[i].lpCode[j + 1], pFunctions[i].lpCode[j + 2]);
							fprintf(fp, "\t\tstart %s;\t\t//%d\n\t\tddw %d;\n", pFunctions[pFunctions[i].lpCode[j + 1]].pszName, pFunctions[i].lpCode[j + 1], pFunctions[i].lpCode[j + 2]);
							break;
						case 0x10062000: //call
							l = 2;
							//fprintf(fp, "\t\tcall;\n\t\tddw %d;\n\t\tddw %d;\n", pFunctions[i].lpCode[j + 1], pFunctions[i].lpCode[j + 2]);
							fprintf(fp, "\t\tcall %s;\t\t//%d\n\t\tddw %d;\n", pFunctions[pFunctions[i].lpCode[j + 1]].pszName, pFunctions[i].lpCode[j + 1], pFunctions[i].lpCode[j + 2]);
							break;
						case 0x10063000: //unk7
							break;
						case 0x10064000: //jmp
							l = 1;
							bFound = false;
							for (z = 0; (z < lnLabels) && (!bFound); z++)
								{
								if (pLabels[z].lOffset == pFunctions[i].lpCode[j + 1]) 
									bFound = true; //if the label points to the destination of our jump
								}
							if (!bFound)
								{
								lnLabels++; //we should never get here since we built our label table first
								pLabels = (_label *)realloc(pLabels, sizeof(_label) * lnLabels);
								pLabels[z].lOffset = pFunctions[i].lpCode[j + 1];
								}
							fprintf(fp, "\t\tjmp Label_%d; //%d\n", z - 1, pFunctions[i].lpCode[j + 1]);
							break;
						case 0x10065000: //ret
							fprintf(fp, "\t\tret;\n");
							break;
						case 0x10066000: //jne
							l = 1;
							bFound = false;
							for (z = 0; (z < lnLabels) && (!bFound); z++)
								{
								if (pLabels[z].lOffset == pFunctions[i].lpCode[j + 1])
									bFound = true;
								}
							if (!bFound)
								{
								lnLabels++;
								pLabels = (_label *)realloc(pLabels, sizeof(_label) * lnLabels);
								pLabels[z].lOffset = pFunctions[i].lpCode[j + 1];
								}
							fprintf(fp, "\t\tjne Label_%d; //%d\n", z - 1, pFunctions[i].lpCode[j + 1]);
							break;
						case 0x10067000: //signal
							fprintf(fp, "\t\tsignal;\n");
							break;
						case 0x10068000: //set signal mask
							fprintf(fp, "\t\tmask;\n");
							break;

						case 0x10071000: //explode
							l = 1;
							fprintf(fp, "\t\texplode %d;\n", pFunctions[i].lpCode[j + 1]);
							break;

						case 0x10082000: //set
							fprintf(fp, "\t\tset;\n");
							break;
						case 0x10083000: //attach
							fprintf(fp, "\t\tattach;\n");
							break;
						case 0x10084000: //drop
							fprintf(fp, "\t\tdrop;\n");
							break;

						case 0x10091000: //cobalt
							fprintf(fp, "\t\t_cobalt;\n");
							break;

						case 0x02800000: //unk8
							break;

						case 0x0: //nop
							fprintf(fp, "\t\tddw 0;\n");
							break;
						
						default:
							break;
						}
					}
				}
			}

		fprintf(fp, "\tendproc;\n");

		if (i < (header.lnFunctions - 1))
			fprintf(fp, "\n");
		}

	fclose(fp);

	return 0;
	}

long decodecob(char *psz)
	{
	char		szBuff[0x1000];
	long		l;

	FILE *fp = fopen(psz, "rb");

	fread(&header, 4, 11, fp);

	pFunctions = new _function[header.lnFunctions];
	pPieces = new _piece[header.lnPieces];

	fseek(fp, header.lOffsetToFunctionNames, SEEK_SET);
	for (long i = 0; i < header.lnFunctions; i++)
		{
		l = 0;
		szBuff[0] = 'a';
		do
			{
			fread(&szBuff[l++], 1, 1, fp);
			}
		while (szBuff[l - 1] != NULL);

		pFunctions[i].pszName = new char [strlen(szBuff) + 1];
		strcpy(pFunctions[i].pszName, szBuff);
		}


	fseek(fp, header.lOffsetToFunctionCodeIndexArray, SEEK_SET);
	for (long j = 0; j < header.lnFunctions; j++)
		fread(&pFunctions[j].lSize, 4, 1, fp);

	for (long z = 0; z < (header.lnFunctions - 1); z++)
		pFunctions[z].lSize = pFunctions[z + 1].lSize - pFunctions[z].lSize;
	pFunctions[header.lnFunctions - 1].lSize = header.lnLongs - pFunctions[header.lnFunctions - 1].lSize;

	long k = 0;

	fseek(fp, header.lOffsetToScriptCode, SEEK_SET);

	while (k < header.lnFunctions)
		{
		pFunctions[k].lpCode = new long [pFunctions[k].lSize];
		fread(pFunctions[k].lpCode, 4, pFunctions[k].lSize, fp);
		k++;			
		}

	//we're at the chr section now
	long m = ftell(fp);
	if (m != header.lOffsetToFunctionCodeIndexArray)
		{
		//we have found preserved symbol names, but this isn't implemented yet so hopefully we won't encounter this
		long lnSymbolNames;
		fread(&lnSymbolNames, 4, 1, fp);
		}

	//read in piece names now
	fseek(fp, header.lOffsetToPieceNameOffsetArray, SEEK_SET);
	long n;
	fread(&n, 4, 1, fp);
	fseek(fp, n, SEEK_SET);
	for (long o = 0; o < header.lnPieces; o++)
		{
		l = 0;
		szBuff[0] = 'a';
		do
			{
			fread(&szBuff[l++], 1, 1, fp);
			}
		while (szBuff[l - 1] != NULL);

		pPieces[o].pszName = new char [strlen(szBuff) + 1];
		strcpy(pPieces[o].pszName, szBuff);
		}

	fclose(fp);

	return 0;
	}